/*
 * MIT License
 *
 * Copyright (c) 2019 norangebit
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package it.norangeb.algorithms.datastructures.dictionary

import arrow.core.None
import arrow.core.Option
import arrow.core.toOption

class RedBlackBST<K : Comparable<K>, V> : OrderedDictionary<K, V> {
    private var root: Option<Node<K, V>> = None

    override fun get(key: K): Option<V> = get(root, key).map { it.value }

    private fun get(
        node: Option<Node<K, V>>,
        key: K
    ): Option<Node<K, V>> = node.flatMap {
        when {
            key == it.key -> node
            key > it.key -> get(it.right, key)
            else -> get(it.left, key)
        }
    }

    override fun set(key: K, value: V) {
        root = set(root, key, value)
    }

    private fun set(
        node: Option<Node<K, V>>,
        key: K,
        value: V
    ): Option<Node<K, V>> = normalize(
        node.fold(
            { Node(key, value) },
            {
                when {
                    key == it.key -> it.clone(value = value)
                    key > it.key -> it.clone(right = set(it.right, key, value))
                    else -> it.clone(left = set(it.left, key, value))
                }
            }
        )
    ).toOption()

    private fun normalize(node: Node<K, V>): Node<K, V> {
        var resultNode = node

        if (isRightRed(resultNode))
            resultNode = rotateLeft(resultNode)

        if (isTwoRedInRow(resultNode))
            resultNode = rotateRight(resultNode)

        if (isBothRed(resultNode))
            resultNode = colorFlip(resultNode)

        return resultNode
    }

    private fun rotateRight(node: Node<K, V>): Node<K, V> = node
        .left.fold(
        { node },
        {
            it.clone(
                color = node.color,
                right = node.clone(color = Color.RED, left = it.right).toOption()
            )
        }
    )

    private fun rotateLeft(node: Node<K, V>): Node<K, V> = node
        .right.fold(
        { node },
        {
            it.clone(
                color = node.color,
                left = node.clone(color = Color.RED, right = it.left).toOption()
            )
        }
    )

    private fun colorFlip(node: Node<K, V>): Node<K, V> = node.clone(
        color = Color.RED,
        left = node.left.map { it.clone(color = Color.BLACK) },
        right = node.right.map { it.clone(color = Color.BLACK) }
    )

    private fun isRightRed(node: Node<K, V>): Boolean = isRed(node.right) &&
            !isRed(node.left)

    private fun isTwoRedInRow(node: Node<K, V>): Boolean = isRed(node.left) &&
            isRed(node.left.flatMap { it.left })

    private fun isBothRed(node: Node<K, V>): Boolean = isRed(node.left) &&
            isRed(node.right)

    private fun isRed(node: Option<Node<K, V>>): Boolean = node
        .fold(
            { false },
            { it.isRed() }
        )

    override fun delete(key: K) {
        TODO("the same as bst")
    }

    override fun contains(key: K): Boolean {
        TODO("the same as bst")
    }

    override fun isEmpty(): Boolean {
        TODO("the same as bst")
    }

    override fun size(): Int {
        TODO("the same as bst")
    }

    override fun max(): Option<K> {
        TODO("the same as bst")
    }

    override fun min(): Option<K> {
        TODO("the same as bst")
    }

    override fun floor(key: K): Option<K> {
        TODO("the same as bst")
    }

    override fun ceiling(key: K): Option<K> {
        TODO("the same as bst")
    }

    override fun select(pos: Int): Option<K> {
        TODO("the same as bst")
    }

    override fun rank(key: K): Int {
        TODO("the same as bst")
    }

    override fun <R> preOrder(transform: (K) -> R) = preOrder(root, transform)

    private fun <R> preOrder(node: Option<Node<K, V>>, transform: (K) -> R) {
        node.map {
            transform(it.key)
            preOrder(it.left, transform)
            preOrder(it.right, transform)
        }
    }

    override fun <R> inOrder(transform: (K) -> R) {
        TODO("the same as bst")
    }

    override fun <R> postOrder(transform: (K) -> R) {
        TODO("the same as bst")
    }

    data class Node<K, V>(
        val key: K,
        val value: V,
        val color: Color = Color.RED,
        val left: Option<Node<K, V>> = None,
        val right: Option<Node<K, V>> = None
    ) {
        fun clone(
            key: K = this.key,
            value: V = this.value,
            color: Color = this.color,
            left: Option<Node<K, V>> = this.left,
            right: Option<Node<K, V>> = this.right
        ): Node<K, V> = Node(key, value, color, left, right)

        fun isRed() = color == Color.RED
    }

    enum class Color {
        RED, BLACK
    }
}