/*
 * MIT License
 *
 * Copyright (c) 2019 norangebit
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

package it.norangeb.algorithms.datastructures.dictionary

import arrow.core.None
import arrow.core.Some
import org.amshove.kluent.`should equal`
import org.spekframework.spek2.Spek
import org.spekframework.spek2.style.gherkin.Feature

object RedBlackBSTTest : Spek({
    Feature("red black tree") {
        Scenario("get and set") {
            val tree by memoized { RedBlackBST<Int, Int>() }

            When("add 1, 3, 4, 6, 8 into tree") {
                tree[1] = 1
                tree[3] = 3
                tree[4] = 4
                tree[6] = 6
                tree[8] = 9
            }

            Then("tree of 1 should equal 1") {
                tree[1] `should equal` Some(1)
            }

            Then("tree of 2 should equal None") {
                tree[2] `should equal` None
            }

            Then("tree of 6 should equal 6") {
                tree[6] `should equal` Some(6)
            }

            Then("tree of 8 should equal 9") {
                tree[8] `should equal` Some(9)
            }

            When("edit 8") {
                tree[8] = 8
            }

            Then("tree of 8 should equal 8") {
                tree[8] `should equal` Some(8)
            }
        }

        Scenario("ordered balance") {
            val tree by memoized { RedBlackBST<Int, Int>() }

            Given("a 1, 2, 3, 4, 5, 6, 7 tree") {
                tree[1] = 1
                tree[2] = 2
                tree[3] = 3
                tree[4] = 4
                tree[5] = 5
                tree[6] = 6
                tree[7] = 7
            }

            val result = mutableListOf<Int>()

            When("compute the preOrder traverse") {
                tree.preOrder { result.add(it) }
            }

            Then("result should equal listOf(4, 2, 1, 3, 6, 5, 7") {
                result `should equal` listOf(4, 2, 1, 3, 6, 5, 7)
            }
        }

        Scenario("reverse balance") {
            val tree by memoized { RedBlackBST<Int, Int>() }

            Given("a 7, 6, 5, 4, 3, 2, 1 tree") {
                tree[7] = 7
                tree[6] = 6
                tree[5] = 5
                tree[4] = 4
                tree[3] = 3
                tree[2] = 2
                tree[1] = 1
            }

            val result = mutableListOf<Int>()

            When("compute the preOrder traverse") {
                tree.preOrder { result.add(it) }
            }

            Then("result should equal listOf(4, 2, 1, 3, 6, 5, 7") {
                result `should equal` listOf(4, 2, 1, 3, 6, 5, 7)
            }
        }

        Scenario("mix balance") {
            val tree by memoized { RedBlackBST<Int, Int>() }

            Given("a 1, 2, 3, 4, 5, 6, 7 tree") {
                tree[6] = 6
                tree[2] = 2
                tree[4] = 4
                tree[1] = 1
                tree[3] = 3
                tree[7] = 7
                tree[5] = 5
            }

            val result = mutableListOf<Int>()

            When("compute the preOrder traverse") {
                tree.preOrder { result.add(it) }
            }

            Then("result should equal listOf(4, 2, 1, 3, 6, 5, 7") {
                result `should equal` listOf(4, 2, 1, 3, 6, 5, 7)
            }

            When("add 8, 9, 10, 12, 13, 14, 15") {
                tree[8] = 8
                tree[9] = 9
                tree[10] = 10
                tree[11] = 11
                tree[12] = 12
                tree[13] = 13
                tree[14] = 14
                tree[15] = 15
            }

            When("compute the preOrder traverse") {
                result.clear()
                tree.preOrder { result.add(it) }
            }

            Then("result should equal listOf(8, 4, 2, 1, 3, 6, 5, 7, 12, 10, 9, 11, 14, 13, 15)") {
                result `should equal` listOf(8, 4, 2, 1, 3, 6, 5, 7, 12, 10, 9, 11, 14, 13, 15)
            }
        }
    }
})